#include <pthread.h>
#include <ncurses.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/select.h>

#include "common.h"
#include "protocol.h"

WINDOW *inputWin;
WINDOW *outputWin;

int sock;
char username[10];
char inBuf[MSGSIZE];
packet_t *packet;
pthread_t thread_packet;

void finish(int retval);
void *waitForPacket();

// Usage: ./client <servername> <username>
int main(int argc, char *argv[]) {
	int test;
	
	if (argc != 3) {
		perror("Wrong amount of arguments!\n");
		exit(1);
	}
	
	sock = connectToServer(argv[1]);
	strcpy(username, argv[2]);
	
	// Init curses
	initscr();				// initialize the curses library
	keypad(stdscr, TRUE);	// enable keyboard mapping
	nonl();					// tell curses not to do NL->CR/NL on output
	cbreak();				// take input chars one at a time, no wait for \n
	echo();					// echo input
	
	// Initialize our windows
	// The input window is just one line at the bottom of the terminal
	// The rest of the terminal is used for output
	outputWin = newwin(LINES - 1, COLS, 0, 0);
	inputWin = newwin(1, COLS, LINES-1, 0);
	
	// Print greeting and move cursor to the input window
	wprintw(outputWin, "Connected to server!\n");
	wrefresh(outputWin);
	wmove(inputWin, 0, 0);
	wrefresh(inputWin);
	
	// Create a thread that immediately calls waitForPacket().
	// This thread does nothing but wait for incoming packets.
	test = pthread_create(&thread_packet, NULL, &waitForPacket, NULL);
	if (test != 0) 
		perror("ERROR pthread");
	
	// Wait for input from terminal
	packet_t pack;
	while(1) {
		strcpy(pack.username, username);
	
		bzero(inBuf, MSGSIZE);
		wgetstr(inputWin, inBuf);
		if (strcmp(inBuf, "exit") == 0)
			break;
		wclear(inputWin);
		wrefresh(inputWin);
		strcpy(pack.msg, inBuf);
		test = write(sock, &pack, sizeof(packet_t));
	} 
	
	// Clean up	
	finish(0);
}

// Connect to a specified server, returns the socket
int connectToServer(char *machine) {
	int test;
	struct hostent *hostp;
	struct sockaddr_in serveraddr;
	// Retrieves info from DNS server.
	if ((hostp = gethostbyname(machine)) != 0) { 
    	sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);    
	
        if (sock < 0) {
       		perror("Could not create socket\n");
       		perror("Please try again later\n");
  			finish(1);
        } else
			printf("Successfully created new socket:%d\n", sock);
	
		bzero(&serveraddr, sizeof(serveraddr));
	  	serveraddr.sin_family = AF_INET;
	  	memcpy(&serveraddr.sin_addr, hostp->h_addr, hostp->h_length);
	  	serveraddr.sin_port = htons(PORT);
	  	test = connect(sock, (struct sockaddr*)&serveraddr, sizeof serveraddr);
          
        if (test != 0) {
			perror("Unable to connect to server\n");
			perror("Please try again later\n");
			finish(1);
        }
	
        return sock;
    } else {
		perror("Unknown machine\n");
		finish(1);
	}
}

// Entrypoint for the packet thread
void *waitForPacket() {
	char buf[PACKSIZE];
	while (1) {
		// Read from socket into our buffer
		bzero(&buf, PACKSIZE);
		read(sock, buf, PACKSIZE);
		packet = (packet_t *) buf;
		
		// Print message in output window and move cursor back to input window
		wprintw(outputWin, "%s: %s\n", packet->username, packet->msg);
		wrefresh(outputWin);
		wmove(inputWin, 0, 0);
		wrefresh(inputWin);
	}
	return NULL;
}

// Clean up and exit program
void finish(int retval) {
	// Send ENDCON packet to server before exiting
	packet_t pack;
	strcpy(pack.username, "ENDCON");
	write(sock, &pack, sizeof(packet_t));
	
	// Close ncurses properly
	endwin();
	
	close(sock);		
	exit(retval);
}
